<?php
/*
simple-gitv - a simple, libre web-based Git project viewer
Copyright (C) 2020

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// STRING PROCESSING

/**
 * If the string does not end in a slash, add one.
 *
 * @param $str a string
 * @return the string with a final slash
 */
function add_trailing_slash($str) {
	if (!empty($str) && substr($str,strlen($str)-1) !== "/") {
		$str = $str . "/";
	}
	return $str;
}

/**
 * Trim beginning until the first slash, if present.
 *
 * @param $str a string
 * @return the substring after the first slash, or the
 * original string if no slash was present
 */
function after_first_slash($str) {
	$slash = strpos($str,"/");
	if ($slash) {
		return substr($str,$slash+1);
	}
	return $str;
}

/**
 * Trim beginning until the last slash, if present.
 *
 * @param $str a string
 * @return the substring after the last slash, or the
 * original string if no slash was present
 */
function after_last_slash($str) {
	$slash = strrpos($str,"/");
	if ($slash) {
		return substr($str,$slash+1);
	}
	return $str;
}

/**
 * Trim end after the first slash, if present.
 *
 * @param $str a string
 * @return the substring before the first slash, or the
 * original string if no slash was present
 */
function before_first_slash($str) {
	$slash = strpos($str,"/");
	if ($slash) {
		return substr($str,0,$slash);
	}
	return $str;
}

/**
 * Trim end after the last slash, if present.
 *
 * @param $str a string
 * @return the substring before the last slash, or the
 * original string if no slash was present
 */
function before_last_slash($str) {
	$slash = strrpos($str,"/");
	if ($slash) {
		return substr($str,0,$slash);
	}
	return $str;
}

/**
 * @param $str a string which may begin with $target
 * @param $target a string
 * @return true if $str begins with $target, false otherwise
 */
function begins_with($str,$target) {
	return substr($str,0,strlen($target)) === $target;
}

/**
 * @param $str a string which may end with $target
 * @param $target a string
 * @return true if $str ends with $target, false otherwise
 */
function ends_with($str,$target) {
	return substr($str,strlen($str) - strlen($target)) === $target;
}

/**
 * Try to find a matching (str1 + str2)
 * in the array of strings.
 *
 * @param $strings an array of strings
 * @param $str1_arr an array of prefixes
 * @param $str2_arr an array of suffixes
 * @return a prefix+suffix concatenated string if it exists in $strings, an empty string otherwise
 */
function find_matching_string($strings,$str1_arr,$str2_arr) {
	foreach ($strings as $str) {
		foreach ($str1_arr as $str1) {
			foreach ($str2_arr as $str2) {
				if (strcasecmp($str,($str1 . $str2)) === 0) {
					return $str;
				}
			}
		}
	}
	return "";
}

/**
 * Attempts to get the project's license from an array
 * of project files, based on common naming conventions
 *
 * It was an intentional choice to use an explicit
 * whitelist of filenames and extensions, instead of
 * checking if it contains "license", "copying", etc. to
 * avoid filenames that happen to contain those words.
 *
 * @param $filenames an array of filenames
 * @return the filename of the license file if found, an empty string otherwise
 */
function get_license($filenames) {
	$license_names = array(
		"COPYING",
		"LICENSE",
		"COPYRIGHT"
		);
	$license_extensions = array(
		"",
		".md",
		".txt"
		);
	return find_matching_string($filenames,$license_names,$license_extensions);
}

/**
 * @param $url the URL following the domain (e.g., if the user goes to https://example.com/project/tree/main/file, $url will be project/tree/main/file)
 * @return page type (specific to simple-gitv)
 */
function get_page_type($url) {
	if (empty($url)) {
		return "home";
	} elseif (!strpos(strip_trailing_slash($url),"/")) {
		return "repo";
	} else {
		$tree_path = "tree/";
		$blob_path = "blob/";
		$raw_path = "raw/";
		$commits_path = "commits";
		$commit_path = "commit/";
		$tags_path = "tags";
		
		if (begins_with(after_first_slash($url),$tree_path)) {
			return "tree";
		} elseif (begins_with(after_first_slash($url),$blob_path)) {
			return "blob";
		} elseif (begins_with(after_first_slash($url),$raw_path)) {
			return "raw";
		} elseif (begins_with(after_first_slash($url),$commits_path)) {
			return "commits";
		} elseif (begins_with(after_first_slash($url),$commit_path)) {
			return "commit";
		} elseif (begins_with(after_first_slash($url),$tags_path)) {
			return "tags";
		}
	}
	return "invalid";
}

/**
 * Attempts to get the project's readme from an array
 * of project files, based on common naming conventions
 *
 * It was an intentional choice to use an explicit
 * whitelist of filenames and extensions, instead of
 * checking if it contains "readme", "read_me", etc. to
 * avoid filenames that happen to contain those words.
 *
 * @param $filenames an array of filenames
 * @return the filename of the readme file if found, an empty string otherwise
 */
function get_readme($filenames) {
	$license_names = array(
		"README",
		"READ_ME"
		);
	$license_extensions = array(
		"",
		".md",
		".txt"
		);
	return find_matching_string($filenames,$license_names,$license_extensions);
}

/**
 * @param $git_dir a git repository directory
 * @return the description of the repository
 */
function get_repo_description($git_dir) {
	return trim(file_get_contents("$git_dir/description"));
}

/**
 * If the string ends in a slash, remove it.
 *
 * @param str a string
 * @return the string without the final slash
 */
function strip_trailing_slash($str) {
	if (!empty($str) && substr($str,strlen($str)-1) === "/") {
		$str = substr($str,0,strlen($str)-1);
	}
	return $str;
}

// GIT METHODS

/**
 * Returns a list of the names of all the branches of
 * a git repo. Note, it does not prettify these names.
 * The active branch will have an asterisk (*) before it.
 *
 * @param $git_dir a git directory
 * @return a list of the names of all the branches
 */
function get_git_branches($git_dir) {
	$res = shell_exec("cd $git_dir && git branch");
	if (!$res) {
		return false;
	}
	return explode("\n",trim($res));
}

/**
 * Finds the git directory of a repo (e.g.,
 * /srv/git/projectname or /srv/git/projectname/.git)
 * 
 * @param $dir a directory
 * @return the git directory, false if not a git directory
 */
function get_git_dir($dir) {
	$git_dir = exec("cd $dir && git rev-parse --git-dir");
	if (! $git_dir) {
		return false;
	} elseif ($git_dir === ".") {
		$git_dir = "$dir";
	} else {
		$git_dir = "$dir/$git_dir";
	}
	return $git_dir;
}

/**
 * @param $git_dir a git directory
 * @return an integer value for the time of the last commit, returns -1 if invalid value
 */
function get_last_commit_time($git_dir) {
	$time = strtotime(trim(shell_exec("cd $git_dir && git log -1 --format=%cd")));
	if (!is_numeric($time)) {
		return -1;
	} else {
		return $time;
	}
}

/**
 * @param $git_dir a git directory
 * @return the name of the active branch
 */
function get_main_git_branch($git_dir) {
	$branches = get_git_branches($git_dir);
	if (! $branches) {
		return false;
	}
	foreach ($branches as $branch) {
		if (begins_with($branch,"*")) {
			return substr($branch,2);
		}
	}
	return false;
}

/**
 * Determines whether a file/directory is actually in the 
 * git repo.
 * 
 * @param $git_dir a git directory
 * @param $path the path in the project to the blob/tree
 * @return true if $path is an actual blob/tree in the project, false otherwise
 */
function is_git_file($git_dir, $path) {
	$result = trim(shell_exec("cd $git_dir && git ls-tree HEAD $path"));
	return !empty($result);
}

/**
 * @param $git_dir a git directory
 * @param $dir the path within the project to check
 * @return an array of items in current directory in git
 * repo, false if none
 */
function list_files($git_dir, $dir = ".") {
	$dir = add_trailing_slash($dir);
	$res = shell_exec("cd $git_dir && git ls-tree --name-only HEAD $dir");
	if (!$res) {
		return false;
	}
	return explode("\n",trim($res));
}

/**
 * @param $git_dir a git directory
 * @param $dir the path within the project to check
 * @return git-ls-tree of directory, false if none
 */
function list_files_verbose($git_dir, $dir = ".") {
	$dir = add_trailing_slash($dir);
	$result = trim(shell_exec("cd $git_dir && git ls-tree HEAD $dir"));
	if (empty($result)) {
		return false;
	}
	
	$result = explode("\n",$result);
	return $result;
}
?>
